
#include "inc.h"

#include "minix.h"
#include "irq_arm.h"
#include "bios_arm.h"
#include "div_arm.h"
#include "gbfs.h"
#include "lib.h"

#include "rand.h"
#include "screen.h"
#include "main.h"

#include "math_arm.h"
#include "task_arm.h"

//---------------------------------------------------------------------------
ST_TASK  Task[TASK_MAX_CNT];
ST_TASK  Dummy;
ST_TASK* pTaskBattle;
ST_TASK* pTaskMyShip;
ST_SPR   Spr;

//math_arm.c
extern s16 m_costbl[];
//screen.c
extern u32 dummy;
//---------------------------------------------------------------------------
IWRAM_CODE void TaskInit()
{
	_Memset((u8*)&Task,  0x00, sizeof(ST_TASK) * TASK_MAX_CNT);
	_Memset((u8*)&Dummy, 0x00, sizeof(ST_TASK) * 1);
}
//---------------------------------------------------------------------------
IWRAM_CODE ST_TASK* TaskSet(void* pFunc)
{
	ST_TASK* p;
	u16 i;

	for(i=0; i<TASK_MAX_CNT; i++)
	{
		if(Task[i].isUse == FALSE)
		{
			p = &Task[i];

			p->isUse = TRUE;
			p->no    = i;
			p->phase = TASK_INIT;
			p->pAct  = pFunc;
			_Memset(p->work, 0, sizeof(p->work));

			//INITs܂
			p->pAct(p);
			return p;
		}
	}

	return &Dummy;
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskExec()
{
	ST_TASK* p;
	u16 i;

	for(i=0; i<TASK_MAX_CNT; i++)
	{
		if(Task[i].isUse == TRUE)
		{
			p = &Task[i];
			p->pAct(p);
		}
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskDel(ST_TASK* p)
{
	p->isUse = FALSE;
}




//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetTitle()
{
	TaskSet(TaskActTitle);
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActTitle(ST_TASK* p)
{
	ST_TASK_WORK_TITLE* w = (ST_TASK_WORK_TITLE*)&p->work[0];
	u16 key;

	switch(p->phase)
	{
	//Sʐݒ
	case TASK_INIT:
		ScreenImgClear();
		Mode3DrawFillBox(0, 0, SCREEN_CX, SCREEN_CY, COLOR_WHITE);
		Mode3DrawFontStr(70, 66, (u8*)"M-KAI PRESENTS", COLOR_BLACK);
		ScreenImgUpdate();

		w->cnt[0] = 100;
		p->phase  = TASK_MAIN;
		break;

	//^Cgʐݒ
	case TASK_MAIN:
		w->cnt[0]--;

		if(w->cnt[0] < 0)
		{
			ScreenImgClear();
			Mode3DrawFillBox(0, 0, SCREEN_CX, SCREEN_CY, COLOR_WHITE);
			Mode3DrawFontStr(42, 44, (u8*)"- MULTI TASK SAMPLE -", COLOR_BLACK); 
			Mode3DrawFontStr(56, 88, (u8*)"- PUSH A BUTTON -",     COLOR_BLACK); 
			ScreenImgUpdate();

			p->phase = TASK_MAIN+1;
		}
		break;

	//L[͑҂
	case TASK_MAIN+1:
		key = KeyGet();

		if(key & KEY_A)
		{
			TaskSetBattle();
			p->phase = TASK_END;
		} 
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}




//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetBattle()
{
	pTaskBattle = TaskSet(TaskActBattle);
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActBattle(ST_TASK* p)
{
	ST_TASK_WORK_BATTLE* w = (ST_TASK_WORK_BATTLE*)&p->work[0];

	switch(p->phase)
	{
	//@̕\
	case TASK_INIT:
		ScreenImgClear();
		Mode3DrawFillBox(0, 0, SCREEN_CX, SCREEN_CY, COLOR_WHITE);
		ScreenImgUpdate();

		TaskSetMyShip();
		p->phase = TASK_MAIN;
		break;

	//G̏o
	case TASK_MAIN:
		w->cnt[0]--;
		w->cnt[1]--;
		w->cnt[2]--;

		if(w->cnt[0] < 0)
		{
			w->cnt[0] = 25;
			TaskSetEnemy1();
		}

		if(w->cnt[1] < 0)
		{
			w->cnt[1] = 150;
			TaskSetEnemy2();
		}

		if(w->cnt[2] < 0)
		{
			w->cnt[2] = 440;
			TaskSetEnemy3();
		}
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}




//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetMyShip()
{
	pTaskMyShip = TaskSet(TaskActMyShip);
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActMyShip(ST_TASK* p)
{
	ST_TASK_WORK_MYSHIP* w = (ST_TASK_WORK_MYSHIP*)&p->work[0];
	u16 key;

	switch(p->phase)
	{
	//
	case TASK_INIT:
		w->XH    = SCR_XH / 2;
		w->YH    = 160 - 32;
		w->speed = 256;

		p->phase = TASK_MAIN;
		break;

	//ړ+e̔
	case TASK_MAIN:
		key   = KeyGet2();
		w->vx = 0;
		w->vy = 0;

		if(key & KEY_RIGHT) w->vx =  w->speed;
		if(key & KEY_LEFT ) w->vx = -w->speed;
		if(key & KEY_UP   ) w->vy = -w->speed;
		if(key & KEY_DOWN ) w->vy =  w->speed;
	    w->X += w->vx;
	    w->Y += w->vy;

		if(w->XH <        12) w->XH = 12;
		if(w->XH > SCR_XH-12) w->XH = SCR_XH-12;
		if(w->YH <        12) w->YH = 12;
		if(w->YH > SCR_YH-12) w->YH = SCR_YH-12;

		if(w->rapidCnt == 0)
		{
			if(key & KEY_A)
			{
				TaskSetMyShot(w->X, w->Y);
				w->rapidCnt = 7;
			}
		}
		else
		{
			w->rapidCnt--;
		}

		SprSet(w->XH-8, w->YH-8, SPR_TILENO_MYSHIP);
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}




//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetMyShot(u16 x, u16 y)
{
	static const u8 shot_dir[] = {192, 192+6, 192-6, 192+12, 192-12};
	ST_TASK* p;
	ST_TASK_WORK_MYSHOT* w;
	u16 i;

	//Ɏˏo܂
	for(i=0; i<5; i++)
	{
		p = TaskSet(TaskActMyShot);
		w = (ST_TASK_WORK_MYSHOT*)&p->work[0];

		w->X     = x;
		w->Y     = y;
		w->speed = 1024;
		w->angle = shot_dir[i];
		w->vx    = SQR_VX(w->angle, w->speed);
		w->vy    = SQR_VY(w->angle, w->speed);
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActMyShot(ST_TASK* p)
{
	ST_TASK_WORK_MYSHOT* w = (ST_TASK_WORK_MYSHOT*)&p->work[0];

	switch(p->phase)
	{
	//́AĂяoōs܂
	case TASK_INIT:
		p->phase = TASK_MAIN;
		break;

	case TASK_MAIN:
	    w->X += w->vx;
	    w->Y += w->vy;

		if( SprIsOut(w->XH, w->YH) == TRUE )
		{
			p->phase = TASK_END;
			break;
		}

		SprSet(w->XH-4, w->YH-4, SPR_TILENO_SHOT1);
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}


//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetEnemy1()
{
	ST_TASK* p;
	ST_TASK_WORK_ENEMY* w;

	p = TaskSet(TaskActEnemy1);
	w = (ST_TASK_WORK_ENEMY*)&p->work[0];

	w->XH     = _UMod(_Rand(), SCR_XH-16) + 8;
	w->YH     = 0;
	w->vx     = 0;
	w->vy     = 256;
	w->cnt[0] = 10;
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActEnemy1(ST_TASK *p)
{
	ST_TASK_WORK_ENEMY* w = (ST_TASK_WORK_ENEMY*)&p->work[0];

	switch(p->phase)
	{
	//́AĂяoōs܂
	case TASK_INIT:
		p->phase = TASK_MAIN;
		break;

	//ړ+e̔
	case TASK_MAIN:
		w->X += w->vx;
		w->Y += w->vy;
		w->cnt[0]--;

		if(w->cnt[0] < 0)
		{
			//                X      Y   SPD  DIR  WAY VWAY
			TaskSetBullet(w->XH, w->YH,  384,   0,   1,   0);
			w->cnt[0] = 32000;
		}

		if( SprIsOut(w->XH, w->YH) == TRUE )
		{
			p->phase = TASK_END;
		}

		//XvCgo^
		SprSet(w->XH-8, w->YH-8, SPR_TILENO_ENEMY1);
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}




//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetEnemy2()
{
	ST_TASK* p;
	ST_TASK_WORK_ENEMY* w;

	p = TaskSet(TaskActEnemy2);
	w = (ST_TASK_WORK_ENEMY*)&p->work[0];

	w->XH     = _UMod(_Rand(), SCR_XH-16) + 8;
	w->YH     = 0;
	w->vx     = 0;
	w->vy     = 900;
	w->cnt[0] = 90;
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActEnemy2(ST_TASK *p)
{
	ST_TASK_WORK_ENEMY* w = (ST_TASK_WORK_ENEMY*)&p->work[0];

	switch(p->phase)
	{
	//́AĂяoōs܂
	case TASK_INIT:
		p->phase = TASK_MAIN;
		break;

	//ړ+e̔
	case TASK_MAIN:
		w->X  += w->vx;
		w->Y  += w->vy;
		w->vy -= 12;
		w->cnt[0]--;

		if(w->cnt[0] < 0)
		{
			//                X        Y   SPD     DIR  WAY  VWAY
			TaskSetBullet(w->XH, w->YH+4,  512, 256+64,   5,   10);
			w->cnt[0] = 2;
		}

		if( SprIsOut(w->XH, w->YH) == TRUE )
		{
			p->phase = TASK_END;
		}

		//XvCgo^
		SprSet(w->XH-8, w->YH-8, SPR_TILENO_ENEMY1);
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}




//---------------------------------------------------------------------------
IWRAM_CODE void TaskSetEnemy3()
{
	ST_TASK* p;
	ST_TASK_WORK_ENEMY* w;

	p = TaskSet(TaskActEnemy3);
	w = (ST_TASK_WORK_ENEMY*)&p->work[0];

	w->XH     = _UMod(_Rand(), SCR_XH-16) + 8;
	w->YH     = 0;
	w->vx     = 0;
	w->vy     = 64;
	w->cnt[0] = 140;
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActEnemy3(ST_TASK *p)
{
	ST_TASK_WORK_ENEMY* w = (ST_TASK_WORK_ENEMY*)&p->work[0];

	switch(p->phase)
	{
	//́AĂяoōs܂
	case TASK_INIT:
		p->phase = TASK_MAIN;
		break;

	//ړ+e̔
	case TASK_MAIN:
		w->X  += w->vx;
		w->Y  += w->vy;
		w->cnt[0]--;

		if(w->cnt[0] < 0)
		{
			//                X      Y   SPD     DIR  WAY VWAY
			TaskSetBullet(w->XH, w->YH,  192, 256+64,   9,   8);
			w->cnt[0] = 70;
		}

		if( SprIsOut(w->XH, w->YH) == TRUE )
		{
			p->phase = TASK_END;
		}

		//XvCgo^
		SprSet(w->XH-8, w->YH-8, SPR_TILENO_ENEMY1);
		break;

	//
	case TASK_END:
		TaskDel(p);
		break;
	}
}




/*=============================================================================
/  Ge
/------------------------------------------------------------------------------
/	Param
/		x, y     = JnWisNZj
/		speed    = exisNZ^t[j
/		angle    = Jn(0`255F_̍@+256FŒ)
/		way      = WAY(0`255)
/		vangle   = A[`̕(0`255)
/============================================================================*/
IWRAM_CODE void TaskSetBullet(u8 x, u8 y, s16 speed, s16 angle, u8 way, u8 vangle)
{
	ST_TASK_WORK_MYSHIP* m = (ST_TASK_WORK_MYSHIP*)&pTaskMyShip->work[0];
	ST_TASK_WORK_BULLET* w;
	ST_TASK*             p;

	u8  ap;			//˕
	u8  va;			//eWay̒SZbg邽߂̕ϐ
	u16 i;

	//0`255̎A_B256ȏ̂Ƃ-256̌ŒɌ܂
	if(angle < 256)
	{
		angle = (GetAngle(m->XH - x, m->YH - y) + angle) & 0xff;
	}
	else
	{
		angle -= 256;
	}

	//WAY̎ASJ悤ɂ܂
	if(way & 0x01)
	{
		angle -= vangle/2;
	}

	ap = angle;
	va = vangle;

	for(i=way; i>=1; i--)
	{
		p = TaskSet(TaskActBullet);
		w = (ST_TASK_WORK_BULLET*)&p->work[0];

		w->vx = SQR_VX(ap, speed);
		w->vy = SQR_VY(ap, speed);
		w->XH  = x;
		w->YH  = y;

		//]
		ap     += va;					//ő256𒴂Ȃ悤ɂ܂
		vangle  = -vangle;				//WayL܂
		va      = -va + vangle;			//̒e̕ɐڂ܂
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void TaskActBullet(ST_TASK *p)
{
	ST_TASK_WORK_BULLET* w = (ST_TASK_WORK_BULLET*)&p->work[0];

	switch(p->phase)
	{
	case TASK_INIT:
		p->phase = TASK_MAIN;
		break;

	case TASK_MAIN:
	    w->X += w->vx;
	    w->Y += w->vy;

		if( SprIsOut(w->XH, w->YH) == TRUE )
		{
			p->phase = TASK_END;
		}

		SprSet(w->XH-4, w->YH-4, SPR_TILENO_SHOT2);
		break;

	case TASK_END:
		TaskDel(p);
		break;
	}
}




//---------------------------------------------------------------------------
IWRAM_CODE void SprInit()
{
	CpuSet((u32)&dummy, (u32)&Spr, (sizeof(ST_SPR) / 2) | CPUSET_16BIT | CPUSET_SRC_FIX);
}
//---------------------------------------------------------------------------
IWRAM_CODE void SprSet(u16 sx, u16 sy, u16 tileNo)
{
	ST_SPR*      p = &Spr; 
	ST_SPR_DATA* w;

	if(p->cnt < SPR_MAX_CNT)
	{
		w = &p->data[p->cnt];

		w->sx     = sx;
		w->sy     = sy;
		w->tileNo = tileNo;

		p->cnt++;
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE void SprDraw()
{
	ST_SPR*      p = &Spr; 
	ST_SPR_DATA* w;
	u16 i;

	for(i=0; i<p->cnt; i++)
	{
		w = &p->data[i];

		SpriteSetSize(i, SP_SIZE_16, SP_SQUARE, SP_COLOR_16);
		SpriteSetChr (i, w->tileNo);
		SpriteMove   (i, w->sx, w->sy);
	}
}
//---------------------------------------------------------------------------
IWRAM_CODE u16 SprGetCnt()
{
	return Spr.cnt;
}
//---------------------------------------------------------------------------
IWRAM_CODE bool SprIsOut(u16 x, u16 y)
{
	return (x > SCR_XH || y > SCR_YH) ? TRUE : FALSE;
}



